/* https://github.com/sphinx-doc/sphinx/issues/1514#issuecomment-742703082 */
/* Newlines (\a) and spaces (\20) before each parameter */
dt > em.sig-param:before {
    content: "\a\20\20\20\20";
    white-space: pre;
}
/* Newline after the last parameter (so the closing bracket is on a new line) */
dt > em.sig-param:last-of-type::after {
    content: "\a";
    white-space: pre;
}
/* For shields */
article.pytorch-article img.shield-badge {
    width: unset;
    margin-top: -18px;
    margin-bottom: 9px;
}
/* Apply code highlight to the whole sentences instead of each word */
code.docutils.literal.notranslate {
    background-color: #f3f4f7;
    border-color: #f3f4f7;
    border-radius: 5px;
    padding: 1px 2px;
}
code.docutils.literal.notranslate span.pre {
    background-color: transparent;
    padding: none;
}
/* Do not overwrite in Tables generated by autosummary */
tr.row-odd code.docutils.literal.notranslate span.pre {
    background-color: transparent;
    border-color: transparent;
}
tr.row-even code.docutils.literal.notranslate {
    background-color: transparent;
    border-color: transparent;
}
/* Fix for Sphinx gallery 0.11
See https://github.com/sphinx-gallery/sphinx-gallery/issues/990
*/
article.pytorch-article .sphx-glr-thumbnails .sphx-glr-thumbcontainer {
    width: unset;
    margin-right: 0;
    margin-left: 0;
}
article.pytorch-article div.section div.wy-table-responsive tbody td {
    width: 50%;
}
/* For customcarditem*/
article.pytorch-article div.tutorials-card div.card-body code {
    color:#d14;
    padding: 2px 4px;
    border-radius: 6px;
    border-top: none;
    border-bottom: none;
    background-color: #AFB8C133;
}
/* C++ doc */
/* Fix the malindentation of `class` header*/
article.pytorch-article .cpp.class dt {
    padding-left: 0.5em;
}
/* keywords like const and explicit, class */
dt.sig.sig-object.cpp > span.k > span.pre {
    color: #c95362;
}

dt.sig.sig-object.cpp > span.n:not(.sig-param) > span.pre {
    color: #9257c6;
}
/* Fix */
dt.sig.sig-object.cpp > span.sig-prename.descclassname > span.n > span.pre {
    color: #6c6c6d;
}
dt.sig.sig-object.cpp > span.sig-prename.descname > span.n > span.pre {
    color: #6c6c6d;
}
