from .impl import (
    TACOTRON2_GRIFFINLIM_CHAR_LJSPEECH,
    TACOTRON2_GRIFFINLIM_PHONE_LJSPEECH,
    TACOTRON2_WAVERNN_CHAR_LJSPEECH,
    TACOTRON2_WAVERNN_PHONE_LJSPEECH,
)
from .interface import Tacotron2TTSBundle


__all__ = [
    "Tacotron2TTSBundle",
    "TACOTRON2_GRIFFINLIM_CHAR_LJSPEECH",
    "TACOTRON2_GRIFFINLIM_PHONE_LJSPEECH",
    "TACOTRON2_WAVERNN_CHAR_LJSPEECH",
    "TACOTRON2_WAVERNN_PHONE_LJSPEECH",
]
